import json
import queue

from core.debug.debugging import log


class PeerEvents:
    def __init__(self, disconnect, listener, peer_state):
        self.disconnect = disconnect
        self.listener = listener

        self.peer_state = peer_state

        self.messages = queue.Queue()
        self.token_requests = queue.Queue()

    def on_event_received(self, event):
        peer_verify_key = self.peer_state.peer_information.get("verify_key")
        my_verify_key = self.peer_state.my_information.get("verify_key")

        if not event or not "type" in event:
            log(f"[{peer_verify_key}] event missing type! {json.dumps(event)}")
            return

        match event["type"]:
            case "message_received":
                self.messages.put([peer_verify_key, event])
                self.listener.emit("message_received", [peer_verify_key, event])
            case "message_sent":
                self.messages.put([my_verify_key, event])
                self.listener.emit("message_sent", [my_verify_key, event])
            case "trusted_token":
                self.token_requests.put([peer_verify_key, event])
            case "disconnect":
                self.disconnect("disconnect gracefully")
                return
            case _:
                self.disconnect("message type not matched")



import pathlib

from PySide6 import QtUiTools, QtCore


def loadUiWidget(file_name, parent=None):
    loader = QtUiTools.QUiLoader()
    uic_file = QtCore.QFile(pathlib.Path("widgets/" + file_name + ".ui"))
    uic_file.open(QtCore.QFile.ReadOnly)
    ui = loader.load(uic_file, parent)
    uic_file.close()
    return ui


from threading import Lock
from operator import lt, le, eq, ne, ge, gt


class PeerManager:
    def __init__(self):
        self.peers = {}
        self.peer_lock = Lock()
        self.ops = {
                "lt": lt,
                "le": le,
                "eq": eq,
                "ne": ne,
                "ge": ge,
                "gt": gt
            }

    def get_peers_of_state(self, comparator, level):
        if comparator not in self.ops:
            raise ValueError(f"Invalid comparator: {comparator}")

        op_func = self.ops[comparator]
        return [peer for peer in self.peers.values() if op_func(peer.peer_state, level)]

    def add_peer(self, peer):
        with self.peer_lock:
            self.peers[peer.verify_key] = peer

    def remove_peer(self, peer_verify_key):
        with self.peer_lock:
            self.peers.pop(peer_verify_key, None)



from nacl.public import PrivateKey


class PeerState:
    def __init__(self, user_manager, my_signing_key):
        self.user_manager = user_manager

        self.my_signing_key = my_signing_key
        self.my_verify_key = self.my_signing_key.verify_key

        self.peer_information = {
            "verify_key": None,
            "public_key": None,
            "addr": None,
            "trusted_token_exists": None
        }

        self.my_information = {
            "verify_key": self.my_verify_key,
            "trusted_token": None
        }

        self.connected = False
        self.trusted = False

    def reload_user_information(self):
        self.my_information = self.user_manager.get_user(self.my_verify_key).serialize()

    def update_peer(self, peer_verify_key, peer_pk, trusted_token_exists):
        self.peer_information = {
            "verify_key": peer_verify_key,
            "public_key": peer_pk,
            "trusted_token_exists": trusted_token_exists
        }
        self.resolve_trusted_state()

    def resolve_trusted_state(self):
        me_trust_peer = bool(self.my_information.get("trusted_token"))
        peer_trust_me = self.peer_information.get("trusted_token_exists", False)
        self.trusted = me_trust_peer or peer_trust_me



import socket
import time

from nacl.signing import SigningKey

import core.globals
from core.networking.discovery_manager import DiscoveryManager


class MockUserManager:
    def __init__(self):
        self.discovered = []
    def set_discovered(self, users):
        print("set_discovered called:", users)
        self.discovered = users
    def on_user_discovered(self, verify_key, addr):
        print(f"discovered user {verify_key.hex()} at {addr}")
        self.discovered.append((verify_key, addr))

um1 = MockUserManager()
um2 = MockUserManager()

vkey1 = SigningKey.generate().verify_key
vkey2 = SigningKey.generate().verify_key

sock1 = socket.socket()
sock1.bind(('', 0))

sock2 = socket.socket()
sock2.bind(('', 0))

dm1 = DiscoveryManager(vkey1, um1, 3)
dm2 = DiscoveryManager(vkey2, um2, 3)

dm1.start_listening()
dm2.start_listening()

def test_discovery_succeeds():
    dm1.start_listening()
    dm2.start_listening()
    dm1.start_broadcast()
    dm2.start_broadcast()
    time.sleep(3)
    assert len(um1.discovered) > 0
    assert len(um2.discovered) > 0



import sys
import threading
import time

from PySide6.QtWidgets import QApplication

from core.node import Node
from notepasser_ui.chat import ChatWindow


class GuiNode(Node):
    def __init__(self, get_trusted_token_input):
        super().__init__(get_trusted_token_input)

        self.app = QApplication(sys.argv)

        self.chat_window = ChatWindow(self)
        self.chat_window.show()
        self.chat_window.user_selected_signal.connect(self.chat_selected)
        self.chat_window.send_message_signal.connect(self.send_message)
        self.chat_window.window_closed_signal.connect(self.disconnect)

        self.cur_peer = None

        self.app.exec()

    def chat_selected(self, user_verify_key):
        user_verify_key = bytes.fromhex(user_verify_key)
        self.cur_peer = self.network_manager.connect_to_peer(user_verify_key)

        if not self.cur_peer: return

        queued_messages = []
        messages_lock = threading.Lock()

        def display_messages():
            nonlocal queued_messages
            while True:
                while not self.cur_peer.peer_events.messages.empty():
                    verify_key, payload = self.cur_peer.peer_events.messages.get()
                    with messages_lock:
                        queued_messages.append(f"[{bytes(verify_key).hex()[:6]}] {payload['message']}")
                time.sleep(0.1)

        msg_thread = threading.Thread(target=display_messages, daemon=True)
        msg_thread.start()

        self.chat_window.load_chat(queued_messages)

        self.cur_peer.listener.subscribe("message_received", self.handle_message_received)
        self.cur_peer.listener.subscribe("message_sent", self.handle_message_received)

    def handle_message_received(self, payload):
        self.chat_window.add_message_to_chat(f"[{bytes(payload[0]).hex()[:6]}] {payload[1]['message']}")

    def send_message(self, message_content):
        if not self.cur_peer: return
        self.cur_peer.peer_connection.send_message(message_content)

    def disconnect(self):
        if not self.cur_peer: return
        self.cur_peer.disconnect()


def main():
    GuiNode(input)

if __name__ == '__main__':
    main()



import json
import threading

from nacl.public import PublicKey, Box
from nacl.signing import VerifyKey

import core.globals
from core.networking.peer_crypto import PeerCrypto


class PeerConnection:
    def __init__(self, disconnect, crypto: PeerCrypto, conn, peer_events):
        self.peer_events = peer_events
        self.disconnect = disconnect

        self.crypto = crypto
        self.conn = conn

    def _send(self, message: dict, encrypt=True):
        data = json.dumps(message).encode()
        if encrypt and self.crypto.box:
            data = self.crypto.box.encrypt(data)

        try:
            self.conn.sendall(data)
        except (ConnectionError, OSError) as e:
            self.disconnect("_send failed: " + str(e))

    def receive(self, decrypt=True, buffer=4096):
        try:
            data = self.conn.recv(buffer)
            if not data:
                return None
            if decrypt and self.crypto.box:
                data = self.crypto.box.decrypt(data)
            return json.loads(data.decode())
        except (ConnectionError, OSError) as e:
            self.disconnect("receive failed: " + str(e))

    def send_message(self, message):
        payload = {"type": "message_received", "message": message}
        self._send(payload)
        self.peer_events.on_event_received({ "type": "message_sent", "message": message })

    def send_disconnect(self):
        payload = {"type": "disconnect"}
        self._send(payload)


import socket
import threading
import time
from concurrent.futures.thread import ThreadPoolExecutor

import core.globals
from core.globals import VERSION
from core.debug.debugging import log
from core.networking.network_manager import NetworkManager
from core.storage.credentials_manager import CredentialsManager
from core.storage.user_manager import UserManager


class DiscoveryManager:
    def __init__(self, credentials_manager: CredentialsManager, user_manager: UserManager, network_manager: NetworkManager, max_broadcast_number):
        self.broadcast_port = core.globals.BROADCAST_PORT
        self.verify_key = credentials_manager.get_signing_key().verify_key
        self.listen_addr = network_manager.listen_sock.getsockname()

        self.user_manager = user_manager
        self.max_broadcast_number = max_broadcast_number

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1) # do NOT allow this for real
        sock.bind((socket.gethostbyname(socket.gethostname()), self.broadcast_port))

        self.sock = sock

        self.broadcast_executor = ThreadPoolExecutor(max_workers=1)
        self.listen_executor = ThreadPoolExecutor(max_workers=1)

        self.replied_to = set()
        self.reply_timeout = 3


    def get_broadcast_string(self, disconnect=False):
        return f"NOTEPASSER|{VERSION}|{bytes(self.verify_key).hex()}|{self.listen_addr[0]}|{self.listen_addr[1]}|{'DISCONNECT' if disconnect else 'CONNECT'}"

    def start_broadcast(self):
        def broadcast():
            log("discovery broadcast started")
            self.user_manager.set_discovered([])
            for i in range(0, self.max_broadcast_number):
                message = self.get_broadcast_string()
                log("sending discovery packet " + message)
                self.sock.sendto(message.encode("utf-8"), ("255.255.255.255", self.broadcast_port))
                time.sleep(1.5)

        self.broadcast_executor.submit(broadcast)

    def start_listening(self):
        def listen():
             log("discovery listen started")
             while True:
                 try:
                     data, addr = self.sock.recvfrom(4096)
                     text = data.decode("utf-8", errors="replace").split("|")
                     if len(text) != 6:
                         log("discovered invalid user")
                         continue
                     prefix, version, peer_verify_key, ip, port, disconnect = text
                     disconnect = disconnect == "DISCONNECT"
                     peer_addr = (ip, int(port))
                     peer_verify_key = bytes.fromhex(peer_verify_key)
                     if prefix != "NOTEPASSER" or version != VERSION:
                         log("different version or irrelevant packet")
                         continue
                     if peer_verify_key == bytes(self.verify_key):
                         log("discovered self")
                         continue
                     log("discovered user " + peer_verify_key.hex() + " with address " + str(peer_addr))

                     self.user_manager.on_user_discovered(peer_verify_key, peer_addr, disconnect)
                     if not disconnect: self.respond_to_discovery_request(peer_verify_key)
                 except socket.timeout:
                     continue
                 except ConnectionResetError:
                     log("windows badness")
                     continue

        self.listen_executor.submit(listen)

    def respond_to_discovery_request(self, peer_verify_key):
        if peer_verify_key in self.replied_to: return
        log("sending discovery packet")
        self.sock.sendto(self.get_broadcast_string().encode("utf-8"), ("255.255.255.255", self.broadcast_port)) # this should really have some kind of limiter
        self.replied_to.add(peer_verify_key)
        self.remove_from_replied_to_after_delay(peer_verify_key)

    def remove_from_replied_to_after_delay(self, peer_verify_key):
        def remove():
            nonlocal peer_verify_key
            time.sleep(self.reply_timeout)
            if not peer_verify_key in self.replied_to: return
            self.replied_to.remove(peer_verify_key)

        threading.Thread(target=remove, daemon=True).start()

    def stop(self):
        self.stop_broadcasting()
        self.listen_executor.shutdown(wait=False)
        message = self.get_broadcast_string(disconnect=True)
        log("shutting down: ", message)
        self.sock.sendto(message.encode('utf-8'), ("255.255.255.255", self.broadcast_port))

    def stop_broadcasting(self): self.broadcast_executor.shutdown(wait=False)



from PySide6.QtCore import Signal
from PySide6.QtGui import QStandardItemModel, QStandardItem, Qt
from PySide6.QtWidgets import QMainWindow, QPushButton, QLineEdit, QListView, QWidget, QVBoxLayout, QStackedLayout, \
    QStackedWidget

import core.globals
from core.debug.debugging import log
from notepasser_ui.utils import loadUiWidget


class ChatWindow(QWidget):
    send_message_signal = Signal(str)
    user_selected_signal = Signal(str)
    window_closed_signal = Signal()

    def __init__(self, node):
        super().__init__()
        self.node = node
        self.window = loadUiWidget("windows/chat_window")

        layout = QVBoxLayout(self)
        layout.addWidget(self.window)
        self.setWindowTitle(f"Notepasser {core.globals.VERSION}")
        self.setLayout(layout)

        self.chat_stack = self.window.findChild(QStackedWidget, "chatStackedWidget")
        self.chat_stack.setCurrentIndex(1)

        self.send_button = self.window.findChild(QPushButton, "sendMessageButton")
        self.message_line_edit = self.window.findChild(QLineEdit, "messageLineEdit")

        self.messages_list_view = self.window.findChild(QListView, "messagesListView")
        self.messages_list_model = QStandardItemModel()
        self.messages_list_view.setModel(self.messages_list_model)

        self.users_list_view = self.window.findChild(QListView, "usersListView")
        self.users_list_model = QStandardItemModel()
        self.users_list_view.setModel(self.users_list_model)
        self.node.user_manager.listener.subscribe("peers_discovered", self.load_users_list)

        self.users_list_view.doubleClicked.connect(self._on_user_selected)
        self.send_button.clicked.connect(self._on_send_button_clicked)

    def load_users_list(self, users_list):
        self.users_list_model.clear()
        if not users_list:
            return

        for user in users_list:
            item = QStandardItem(bytes(user).hex())
            self.users_list_model.appendRow(item)

    def unload_chat(self):
        self.chat_stack.setCurrentIndex(1)

    def load_chat(self, queued_messages):
        self.chat_stack.setCurrentIndex(0)
        for message in queued_messages:
            item = QStandardItem(message)
            self.messages_list_model.appendRow(item)

    def add_message_to_chat(self, message):
        self.chat_stack.setCurrentIndex(0)

        item = QStandardItem(message)
        self.messages_list_model.appendRow(item)

    def _on_user_selected(self):
        selected_indexes = self.users_list_view.selectedIndexes()[0]
        user = self.users_list_model.data(selected_indexes, Qt.DisplayRole) if selected_indexes else None
        self.user_selected_signal.emit(user)

    def _on_send_button_clicked(self):
        message = self.message_line_edit.text()
        if message:
            self.send_message_signal.emit(message)
            self.message_line_edit.clear()

    def closeEvent(self, event, /):
        self.window_closed_signal.emit()
        event.accept()


from core.networking.discovery_manager import DiscoveryManager
from core.networking.network_manager import NetworkManager
from core.storage.credentials_manager import CredentialsManager
from core.storage.storage_manager import StorageManager
from core.storage.user_manager import UserManager


class Node:
    def __init__(self, get_trusted_token_input, config_dir=".notepasser"):
        self.storage_manager = StorageManager(config_dir)
        self.user_manager = UserManager(self.storage_manager)
        self.credentials_manager = CredentialsManager(self.storage_manager)

        self.network_manager = NetworkManager(self.credentials_manager, self.user_manager, get_trusted_token_input)
        self.discovery_manager = DiscoveryManager(self.credentials_manager, self.user_manager, self.network_manager, 3)

        self.discovery_manager.start_listening()
        self.discovery_manager.start_broadcast()



import json
import random
import threading
from socket import socket

from nacl.public import PublicKey
from nacl.signing import SigningKey, VerifyKey

from core.globals import running
from core.networking.peer_connection import PeerConnection
from core.networking.peer_crypto import PeerCrypto
from core.networking.peer_events import PeerEvents
from core.networking.peer_state import PeerState
from core.storage.user_manager import UserManager
from core.debug.debugging import log
from core.utils.listener import Listener


class Peer:
    def __init__(self, network_manager, user_manager: UserManager, conn, addr, my_sign: SigningKey, get_trusted_token):
        self.network_manager = network_manager
        self.user_manager = user_manager
        self.conn = conn
        self.addr = addr

        self.listener = Listener(["connected", "message_received", "message_sent"])

        self.peer_state = PeerState(user_manager, my_sign)
        self.peer_crypto = PeerCrypto(self.peer_state, get_trusted_token)
        self.peer_events = PeerEvents(self.disconnect, self.listener, self.peer_state)
        self.peer_connection = PeerConnection(self.disconnect, self.peer_crypto, self.conn, self.peer_events)

        self._disconnected = False
        log("new peer created with random debug number of: " + str(random.randrange(1000, 9999)))
        self.start()

    def start(self):
        threading.Thread(target=self.listen_for_connection_information, daemon=True).start()

        self.conn.sendall(self._get_connection_string())

    def _get_connection_string(self):
        payload = {
            "type": "connection".encode().hex(),
            "encryption_key": bytes(self.peer_crypto.my_pk).hex(),
            "signature": self.peer_crypto.my_signing_key.sign(bytes(self.peer_crypto.my_pk)).signature.hex(),
            "verify_key": bytes(self.peer_crypto.my_verify_key).hex(),
            "trusted_token_exists": bool(self.peer_state.my_information.get("trusted_token"))
        }

        return json.dumps(payload).encode()

    def listen_for_connection_information(self):
        while running and not self._disconnected and not self.peer_state.peer_information.get("verify_key"):
            data = self.peer_connection.receive()

            peer_encryption_key = bytes.fromhex(data["encryption_key"])
            peer_signature = bytes.fromhex(data["signature"])
            peer_verify_key = VerifyKey(bytes.fromhex(data["verify_key"]))
            peer_trusted_token_exists = bool(data["trusted_token_exists"])

            peer_verify_key.verify(peer_encryption_key, peer_signature)

            peer_pk = PublicKey(peer_encryption_key)
            self.peer_crypto.set_my_box(peer_pk)
            self.peer_state.update_peer(peer_verify_key, peer_pk, peer_trusted_token_exists)

            self.peer_state.connected = True

            self.peer_state.reload_user_information()
            self.peer_state.resolve_trusted_state()

        log(self.peer_state.peer_information.get("verify_key"))

        threading.Thread(target=self.listen_for_events, daemon=True).start()

    def listen_for_events(self):
        log("listening for events")
        log(self.conn)

        while running and not self._disconnected:
            event = self.peer_connection.receive()
            log(event)
            self.peer_events.on_event_received(event)

        self.disconnect("listen_for_events: no longer running")

    def disconnect(self, reason=None):
        log("disconnecting for : " + str(reason))

        if self._disconnected:
            return
        self._disconnected = True

        # send disconnect only if socket is still valid
        try:
            if self.conn and isinstance(self.conn, socket):
                self.peer_connection.send_disconnect()
        except OSError:
            pass

        try:
            if self.conn and isinstance(self.conn, socket):
                self.conn.close()
        except OSError:
            pass

        self.network_manager.remove_peer(self, reason)



import socket
import threading
import time

import core.globals
from core.networking.discovery_manager import DiscoveryManager
from core.networking.network_manager import NetworkManager
from core.storage.credentials_manager import CredentialsManager
from core.storage.storage_manager import StorageManager
from core.storage.user_manager import UserManager
from core.debug.debugging import log
from core.globals import running

class Colors:
    HEADER = "\033[95m"
    ORANGE = "\033[38;5;208m"
    CYAN = "\033[96m"
    BROWN = "\033[38;5;94m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    RESET = "\033[0m"
    BOLD = "\033[1m"

def print_banner():
    print(f"{Colors.ORANGE}{'='*50}")
    print(f"""{Colors.YELLOW}
⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢀⣤⠞⠛⣦⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀
⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢀⡴⠋⢀⠤⡀⠸⡇⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀
⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢀⡴⠋⢀⡄⢧⢉⢦⠀⢻⡀⠀⠀⠀⠀⠀⣀⣤⠴⠶⣄⠀⠀
⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⣠⣞⣁⣴⣑⣪⣐⣮⣘⣀⣸⡇⠀⠀⣠⡴⠚⠉⡀⠤⡄⢸⡆⠀
⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⣠⠶⠛⢉⡁⢀⣀⠀⠀⠀⠀⠀⠉⠉⠙⠓⠾⣭⣀⠰⡍⢓⠥⠇⠈⡇⠀
⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⣸⠃⠀⢐⢇⠼⣘⠔⣣⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠙⠃⢜⠥⠪⡇⢀⡇⠀
⠀⠀⠀⠀⠀⠀⠀⠀⢰⡟⣦⣄⡀⠀⠀⢿⠀⠀⠩⡌⠖⣡⢚⠊⠀⣰⡷⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠙⣧⠃⢸⠇⠀
⠀⠀⠀⠀⠀⠀⠀⠀⢸⡇⣿⣾⣿⣶⣄⠘⣇⠀⠀⠈⠁⢉⣤⡀⠀⠈⠀⠀⠀⣴⡷⠀⠀⠀⠀⠀⠀⠀⠈⢧⠟⠀⠀
⠀⠀⠀⠀⠀⠀⠀⠀⠐⣯⢺⣿⣿⢿⣿⣷⣜⢧⡀⠀⣰⠟⣼⣇⠀⠀⠀⠀⠀⠀⠀⠀⢠⠲⢍⢏⡂⡀⠀⠘⡇⠀⠀
⠀⠀⠀⠀⠀⠀⠀⠀⠀⢹⡎⣿⡿⣟⣿⣳⣿⣦⠙⣷⢏⣾⣿⣿⠀⠀⠀⠀⠀⠀⠀⠀⢠⢃⡎⡬⢡⣱⠀⠀⡿⠀⠀
⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢷⡹⣿⣿⣻⣿⡽⣿⣧⡏⣽⣿⡿⣿⠀⢀⣤⢞⣿⠀⠀⠀⠀⠃⠜⠆⠗⠁⠀⣰⠇⠀⠀
⠀⠀⠀⠀⣀⣄⣀⣀⠀⠀⠘⣷⢻⣷⣻⣟⣿⡿⣽⣧⣿⣿⣽⣿⡴⢏⣵⣿⣿⣀⠀⠀⠀⠀⠀⠀⠀⣀⡴⠋⠀⠀⠀
⠀⠀⠀⠀⢿⡙⣭⣭⣯⣛⣶⢮⣷⣻⣿⣻⡿⣽⢿⣾⣿⣷⣿⣿⡟⠛⠛⠛⠧⣌⠙⠛⣗⠒⠒⠛⠋⠁⠀⠀⠀⠀⠀
⠀⠀⠀⠀⠈⢷⡽⣿⣻⣿⣻⣿⢿⡿⣟⣯⣿⣎⢿⣟⣯⣽⣿⣾⡀⠀⠀⠀⠀⠀⠀⠀⢻⡀⠀⠀⠀⠀⠀⠀⠀⠀⠀
⠀⠀⠀⠀⠀⠀⠙⢮⣛⢷⡿⣾⣯⢿⣿⢿⣯⠿⣯⢻⣾⣿⣷⣿⣧⡀⠀⠀⠀⠀⠀⠀⢸⡇⠀⠀⠀⠀⠀⢀⣴⠶⣦
⠀⠀⠀⠀⠀⠀⠀⠀⠉⠛⠶⢭⣽⣿⣿⡿⣟⣿⣶⣭⠛⣾⣿⣿⣿⣽⣲⠆⠀⠀⠀⠀⢸⠃⠀⠀⠀⠀⣰⠟⠁⠀⣸
⠀⠀⠀⠀⠀⠀⠀⠀⢀⣤⢞⣯⣿⣻⣽⣿⣯⣿⣿⡿⣿⣿⣿⣾⣿⠿⠿⠃⠀⠀⠀⢀⡟⠀⠀⢀⣠⠞⠁⠀⠀⣰⠏
⠀⠀⠀⠀⠀⠀⢀⣴⢏⣾⣿⣿⣯⣿⣿⣷⣻⣿⣾⣽⠟⣿⣧⠀⠀⠀⠀⠀⠀⠀⣠⣞⣠⣤⠶⠋⠁⠀⠀⢀⡼⠋⠀
⠀⠀⠀⠀⠀⠀⠈⠙⠛⠛⠛⠛⠉⠉⢰⣏⣿⡿⢿⣅⠀⢸⣿⡄⠀⠀⠀⠀⣠⡴⠋⠉⠁⠀⠀⠀⣀⣤⠶⠋⠀⠀⠀
⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠘⠋⠁⠀⠀⠉⠓⠾⠿⠤⠤⠶⠖⠻⠷⠶⠤⠶⠶⠖⠚⠋⠉⠀⠀⠀⠀⠀⠀
⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀art from https://emojicombos.com/autumn-ascii-art⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀
⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀""")
    print(f"{Colors.BOLD}{Colors.RED}        Notepasser {core.globals.VERSION} FALL UPDATE")
    print(f"{Colors.ORANGE}{'='*50}{Colors.RESET}\n")

def main(storage_location=None):
    storage = StorageManager(storage_location) if storage_location else StorageManager()
    user_manager = UserManager(storage)
    credentials = CredentialsManager(storage)

    network = NetworkManager(credentials, user_manager, input)
    discovery = DiscoveryManager(
        credentials_manager=credentials,
        user_manager=user_manager,
        network_manager=network,
        max_broadcast_number=3
    )

    discovery.start_listening()
    discovery.start_broadcast()
    print(f"{Colors.BROWN}[DISCOVERY] Listening and broadcasting started...{Colors.RESET}")

    print_banner()

    try:
        while running:
            print(f"{Colors.YELLOW}\nCOMMANDS:{Colors.RESET}")
            print(f"{Colors.ORANGE}- discover{Colors.RESET} - Show discovered peers")
            print(f"{Colors.ORANGE}- connect{Colors.RESET} - Connect to a peer by index")
            print(f"{Colors.ORANGE}- exit{Colors.RESET} - Quit program")
            print(f"{Colors.CYAN}{'-'*30}{Colors.RESET}")

            cmd = input(f"{Colors.BOLD}> {Colors.RESET}").strip().lower()

            if cmd == "exit":
                print(f"{Colors.RED}[EXIT] Shutting down...{Colors.RESET}")
                break

            elif cmd == "discover":
                peers = user_manager.discovered
                if not peers:
                    print(f"{Colors.YELLOW}[DISCOVERY] No peers found yet.{Colors.RESET}")
                else:
                    print(f"{Colors.BROWN}[DISCOVERY] Found peers:{Colors.RESET}")
                    for i, p in enumerate(peers):
                        print(f"  [{i}] {p}")

            elif cmd == "connect":
                peers = user_manager.discovered
                if not peers:
                    print(f"{Colors.YELLOW}[CONNECT] No peers discovered.{Colors.RESET}")
                    continue

                print(f"{Colors.BROWN}Select peer index:{Colors.RESET}")
                for i, p in enumerate(peers):
                    print(f"  [{i}] {p}")
                try:
                    idx = int(input(f"{Colors.BOLD}> {Colors.RESET}"))
                    target = peers[idx]
                except (ValueError, IndexError):
                    print(f"{Colors.RED}[ERROR] Invalid index.{Colors.RESET}")
                    continue

                peer = network.connect_to_peer(target)
                if not peer:
                    print(f"{Colors.RED}[CONNECT] Failed to connect.{Colors.RESET}")
                    continue

                print(f"{Colors.BROWN}[CONNECTED] Messaging session started with {target}.{Colors.RESET}")
                print(f"{Colors.YELLOW}Type '|exit|' to leave chat{Colors.RESET}\n")

                chat_active = True

                def display_messages():
                    while running and chat_active:
                        while not peer.peer_events.messages.empty():
                            addr, message = peer.peer_events.messages.get()
                            print(f"\n{Colors.CYAN}[{bytes(addr).hex()[:6]}]:{Colors.RESET} {message['message']}")
                            print(f"{Colors.BOLD}> {Colors.RESET}", end="", flush=True)
                        time.sleep(0.1)

                msg_thread = threading.Thread(target=display_messages, daemon=True)
                msg_thread.start()

                while running:
                    msg = input(f"{Colors.BOLD}> {Colors.RESET}")
                    if msg == "|exit|":
                        print(f"{Colors.RED}[CHAT] Ending session...{Colors.RESET}")
                        chat_active = False
                        peer.disconnect()
                        break
                    elif msg.strip():
                        peer.peer_connection.send_message(msg)

            else:
                print(f"{Colors.RED}[ERROR] Unknown command.{Colors.RESET}")

    except KeyboardInterrupt:
        print(f"\n{Colors.RED}[EXIT] Interrupted by user.{Colors.RESET}")
    finally:
        print(f"{Colors.YELLOW}[CLEANUP] Disconnecting all peers...{Colors.RESET}")
        for peer in list(network.peers.values()):
            peer.disconnect()
        discovery.stop()
        print(f"{Colors.BROWN}[EXIT] Shutdown complete.{Colors.RESET}")



import threading

from core.debug.debugging import log


class Listener:
    def __init__(self, event_names):
        self.lock = threading.RLock()

        self.subscriptions = {}

        for event_name in event_names:
            self.subscriptions[event_name] = []

    def subscribe(self, event_name, callback):
        with self.lock:
            if event_name not in self.subscriptions:
                log(f"event {event_name} does not exist")
                return

            log(f"subscribed to {event_name}")
            self.subscriptions[event_name].append(callback)

    def unsubscribe(self, event_name, callback=None):
        with self.lock:
            if not callback and event_name:
                self.subscriptions[event_name] = []
                log(f"cleared {event_name}")
            elif event_name and callback in self.subscriptions[event_name]:
                self.subscriptions[event_name].remove(callback)
                log(f"unsubscribed to {callback.__name__} in {event_name}")

    def emit(self, event_name, *args, **kwargs):
        with self.lock:
            callbacks = list(self.subscriptions.get(event_name, []))

        for callback in callbacks:
            callback(*args, **kwargs)
        log(f"emitted {event_name}")
